#include <assert.h>
#include <stdlib.h>
#include <math.h>
#include "config.h"
#include "randperm.h"
#include "turbo.h"

#define ENCODER_MEM         4
#define ENCODER_STATES      (1<<(ENCODER_MEM))
#define ENCODER_STATE_MASK  (ENCODER_STATES-1)

#define logsum(x,y)         logsum_sqr(x, y)

typedef UINT16 STATE;

static STATE trellis[ENCODER_STATES][2];
static STATE invtrel[ENCODER_STATES][2];
static BIT output1[ENCODER_STATES][2];

static INT32 RANDPERM_PRIM;

static void make_trellis ()
{
    STATE st;

    assert(ENCODER_MEM < 16);

    for (st = 0; st < ENCODER_STATES; st++) {
        STATE sumst, newst, bit;

        /* st >> i is D^(i+1) */
        sumst = 1 & (st ^ (st >> 1) ^ (st >> 2) ^ (st >> 3));
        newst = ((st << 1) & ENCODER_STATE_MASK) | sumst;
        sumst ^= (st >> 3);  /* 1 & (st >> 3) */

        for (bit = 0; bit < 2; bit++) {
            trellis[st][bit] = newst ^ bit;
            invtrel[newst ^ bit][bit] = st;
            output1[st][bit] = bit ^ sumst;
        }
    }
}


/** x = encode(u);
 *  where u is a row vector containing information
 *  we will try, after trunc_len dummy bits, to ensure the state be 0.
 */
void turbo_encode (BIT *u, BIT **x, UINT32 len, UINT32 trunc_len)
{
    STATE st1, st2;
    BIT *u0, *x1, *x2;
    UINT32 i;

    assert(trellis[0][1] != 0);
    assert(RANDPERM_PRIM == len+1);

    st1 = st2 = 0;
    u0 = x[0]; x1 = x[1];
#ifndef TURBO_ORIG
    x2 = x[2];
#endif

    randperm_reset();
    for (i = 0; i < len+trunc_len; i++) {
        BIT ui, up;

        if (i < len) {
            ui = u[i];
            up = u[randperm_next() - 1];
        }
        else {   /* trailing bits */
            ui = 1 & (st1 ^ (st1 >> 1) ^ (st1 >> 2) ^ (st1 >> 3));
            up = 1 & (st2 ^ (st2 >> 1) ^ (st2 >> 2) ^ (st2 >> 3));
        }

        *u0++ = ui;
#ifdef TURBO_ORIG
        if (i & 1)
            *x1++ = output1[st2][up];
        else
            *x1++ = output1[st1][ui];
#else
        *x1++ = output1[st1][ui];
        *x2++ = output1[st2][up];
#endif
        st1 = trellis[st1][ui];
        st2 = trellis[st2][up];
    }

    assert(st1 == 0);
    assert(st2 == 0);
}


static REAL logsum_max (REAL x, REAL y)
{
    if (x > y) return x;
    else return y;
}


static REAL logsum_sqr (REAL x, REAL y)
{
/*
#define Y1  0.58913471818505
#define Y2  0.38406926874035
#define Y3  0.19232813672731
#define Y4  0.03913502618594

#define D1  0.46716125369449
#define D2  1.09654736705310
#define D3  2.09805791875402
*/
#define Y1  0.58279078244958
#define Y2  0.36608274577063
#define Y3  0.16720102566285
#define Y4  0.01678202190027

#define D1  0.49904732573698
#define D2  1.18556717090270
#define D3  2.33971050027289

    REAL m, d;
    m = logsum_max(x, y);
    d = fabs(x - y);

    if (d > D3) return m + Y4;
    else if (d > D2) return m + Y3;
    else if (d > D1) return m + Y2;
    else return m + Y1;
}


static REAL logsum_real (REAL x, REAL y)
{
    if (x < REAL_NEG_INFINITY / 2)
        return y;
    else if (y < REAL_NEG_INFINITY / 2)
        return x;

    return logsum_max(x, y) + log(1 + exp(-fabs(x - y)));
}

/* beta is a big matrix -- in order not to alloc/free it again and again,
 * I put the alloc/free operation to decode()
 */
static void turbo_bcjr (REAL *v, REAL *y, REAL *pi, REAL *pi2,
                        REAL sigma, UINT32 len, UINT32 trunc_len,
                        REAL **beta)
{
    /* calculate upward message lambda */
    /* edge weight: p(y|x) * pi(u) */
    /* for truncation bits: pi(u) = 1/2, or LLR_pi(u) = 0 */

    REAL snr = 1 / sigma / sigma;
    REAL alpha[ENCODER_STATES];
    /* weig0 weig1: alpha matrix with different u */
    REAL weig0[ENCODER_STATES], weig1[ENCODER_STATES];
    STATE s, s0, s1;
    UINT32 t, tlen = len + trunc_len;  /* total length */

    assert(tlen > len);

    beta[tlen][0] = 0;
    for (s = 1; s < ENCODER_STATES; s++)
        beta[tlen][s] = REAL_NEG_INFINITY;

    for (t = tlen; t > 0; t--) {
        REAL d1, w0, w1;

        d1 = snr * y[t-1];
        for (s = 0; s < ENCODER_STATES; s++) {
            s0 = trellis[s][0];
            s1 = trellis[s][1];
            assert(s0 >= 0 && s0 < ENCODER_STATES);
            assert(s1 >= 0 && s1 < ENCODER_STATES);
            assert(output1[s][0] + output1[s][1] == 1);

            if (output1[s][0]) {
                w0 = pi[t-1] - d1;
                w1 = d1;
            }
            else {
                w0 = pi[t-1] + d1;
                w1 = - d1;
            }
#ifdef TURBO_BCJR
            w0 += 2 * snr * v[t-1];
#endif
            beta[t-1][s] = logsum(beta[t][s0] + w0, beta[t][s1] + w1);
        }
    }


    alpha[0] = 0;
    for (s = 1; s < ENCODER_STATES; s++)
        alpha[s] = REAL_NEG_INFINITY;

    // Calculate alpha & pi2 (in fact, pi2+pi)
    for (t = 0; t < len; t++) {
        REAL pu0, pu1, d1;
        pu0 = pu1 = REAL_NEG_INFINITY;

        d1 = snr * y[t];
        for (s = 0; s < ENCODER_STATES; s++) {
            s0 = trellis[s][0];
            s1 = trellis[s][1];

            //d1 = snr * y[t] * (1-2*output1[s][0]);
            assert(output1[s][0] + output1[s][1] == 1);
            if (output1[s][0]) {
                weig0[s] = alpha[s] + pi[t] - d1;
                weig1[s] = alpha[s] + d1;
            }
            else {
                weig0[s] = alpha[s] + pi[t] + d1;
                weig1[s] = alpha[s] - d1;
            }
#ifdef TURBO_BCJR
            weig0[s] += 2 * snr * v[t];
#endif
            pu0 = logsum(pu0, weig0[s] + beta[t+1][s0]);
            pu1 = logsum(pu1, weig1[s] + beta[t+1][s1]);
        }
#ifdef TURBO_BCJR
        pi2[t] = pu0 - pu1 - pi[t];
#else
        pi2[t] = pu0 - pu1 + snr * (v[t]+v[t]) - pi[t];
#endif

        for (s = 0; s < ENCODER_STATES; s++) {
            s0 = invtrel[s][0];
            s1 = invtrel[s][1];

            alpha[s] = logsum(weig0[s0], weig1[s1]);
        }
    }
}


void turbo_decode_sub (REAL **y, REAL *pi, REAL *LLR, UINT16 iters,
                       REAL sigma, UINT32 len, UINT32 trunc_len, int stage)
{
    REAL *u1;
    static REAL *u2, *tmp, **beta;
    UINT32 t, tlen = len + trunc_len;
    UINT16 ite;

    u1 = y[0];
    if (stage == 0) {
        u2 = (REAL*) malloc(sizeof(REAL) * tlen);
        randperm_reset();
        for (t = 0; t < len; t++)
            u2[t] = u1[randperm_next()-1];
        for (t = len; t < tlen; t++)
            u2[t] = 0;      /* we do not know the trailing bits */

        tmp = (REAL*) malloc(sizeof(REAL) * len);

        beta = (REAL**) malloc(sizeof(REAL*) * (tlen + 1));
        for (t = 0; t < tlen + 1; t++) {
            beta[t] = (REAL*) malloc(sizeof(REAL) * ENCODER_STATES);
            assert(beta[t] != NULL); /* ??? */
        }
        return;
    }
    assert(u1 != NULL);
    assert(u2 != NULL);
    assert(tmp != NULL);
    assert(beta != NULL);

    if (stage == 2) {
        for (t = 0; t < tlen + 1; t++)
            free(beta[t]);
        free(beta);
        free(u2);
        free(tmp);
        return;
    }

    for (ite = 0; ite < iters; ite++) {
        turbo_bcjr(u1, y[1], pi, tmp, sigma, len, trunc_len, beta);

        randperm_reset();
        for (t = 0; t < len; t++)
            pi[t] = tmp[randperm_next()-1];

        turbo_bcjr(u2, y[2], pi, tmp, sigma, len, trunc_len, beta);

        if (ite == iters-1) {
            randperm_reset();
            for (t = 0; t < len; t++)
                LLR[randperm_next()-1] = pi[t] + tmp[t];
        }

        randperm_reset();
        for (t = 0; t < len; t++)
            pi[randperm_next()-1] = tmp[t];
    }
}


void turbo_decode (REAL **y, BIT *u, UINT16 iters,
                   REAL sigma, UINT32 len, UINT32 trunc_len)
{
    REAL *pi, *LLR;
    UINT32 t, tlen = len + trunc_len;

    LLR = (REAL*) malloc(sizeof(REAL) * len);
    pi = (REAL*) malloc(sizeof(REAL) * tlen);
    for (t = 0; t < tlen; t++) pi[t] = 0;

#ifdef TURBO_ORIG
    for (t = 1; t < tlen; t+=2) {
        y[2][t-1] = 0;
        y[2][t] = y[1][t];
        y[1][t] = 0;
    }
    if (tlen & 1)
        y[2][tlen-1] = 0;
#endif

    turbo_decode_sub (y, pi, LLR, iters, sigma, len, trunc_len, 0);
    turbo_decode_sub (y, pi, LLR, iters, sigma, len, trunc_len, 1);
    turbo_decode_sub (y, pi, LLR, iters, sigma, len, trunc_len, 2);

    for (t = 0; t < len; t++)
        u[t] = (LLR[t] < 0);

    free(pi);
    free(LLR);
}

void turbo_init (INT32 base, INT32 seed, INT32 prim)
{
    RANDPERM_PRIM = prim;
    randperm_init(base, seed, prim);
    make_trellis();
}
